# -*- coding: utf-8 -*-
"""
    Efesto

    Copyright (C) 2016 Jacopo Cascioli

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import falcon

from .Base import config
from .Models import (AccessRules, EternalTokens, Fields, Types, Users,
                     make_model)
from .Resources import (RootResource, TokensResource, make_collection,
                        make_resource)
from .Version import __version__


if config.parser.getboolean('main', 'installed'):
    app = falcon.API()
    root_message = 'Running efesto %s' % (__version__)
    root_data = {'message': root_message}
    app.add_route('/', RootResource(root_data))

    for i in [['/users', Users], ['/types', Types], ['/fields', Fields],
              ['/rules', AccessRules], ['/tokens', EternalTokens]]:
        collection = make_collection(i[1])()
        resource = make_resource(i[1])()
        app.add_route(i[0], collection)
        app.add_route('%s/{id}' % i[0], resource)

    custom_types = Types.select().where(Types.enabled == True)
    for custom_type in custom_types:
        model = make_model(custom_type)
        model_name = getattr(model._meta, 'db_table')
        collection = make_collection(model)()
        resource = make_resource(model)()
        app.add_route('/%s' % (model_name), collection)
        app.add_route('/%s/{id}' % (model_name), resource)

    app.add_route('/auth', TokensResource())

    def error_serializer(req, exception):
        preferred = 'application/json'
        representation = exception.to_json()
        return (preferred, representation)
    app.set_error_serializer(error_serializer)



# -*- coding: utf-8 -*-
"""
    The Efesto Blueprints module.

    Copyright (C) 2016 Jacopo Cascioli

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
from configparser import ConfigParser

from .Models import Fields, Types


def load_blueprint(blueprint):
    """
    Loads a blueprint in to the database from a blueprint file.
    """
    path = os.path.join(os.getcwd(), blueprint)
    parser = ConfigParser()
    parser.read(path)

    types = []
    for section in parser.sections():
        if '.' not in section:
            types.append(section)

    for type in types:
        new_type = Types(name=type, enabled=0)
        new_type.save()
        fields = parser.get(type, 'fields').split(',')
        for field in fields:
            field_name = field.strip()
            try:
                field_section = '{}.{}'.format(type, field_name)
                field_type = parser.get(field_section, 'type')
            except:
                field_type = 'string'
            new_field = Fields(name=field_name, type=new_type.id,
                               field_type=field_type)
            new_field.save()
        new_type.enabled = 1
        new_type.save()


def dump_blueprint(blueprint_file):
    """
    Dumps a blueprint, creating a blueprint file from the database data.
    """
    parser = ConfigParser()
    types = Types.select()
    for type in types:
        parser.add_section(type.name)
        fields = Fields.select().where(Fields.type == type.id)
        fields_list = []
        for field in fields:
            if field.field_type != 'string':
                field_section = '{}.{}'.format(type.name, field.name)
                parser.add_section(field_section)
                parser.set(field_section, 'type', field.field_type)
            fields_list.append(field.name)
        parser.set(type.name, 'fields', ', '.join(fields_list))

    with open(blueprint_file, 'w') as blueprint:
        parser.write(blueprint)



# -*- coding: utf-8 -*-
"""
    Efesto travis script.

    This script sets up Efesto for travis, creating tables and settings.
"""
import sys

from efesto.Base import config


sys.path.insert(0, '')


def create_travis_config():
    config.parser.set('db', 'name', 'test')
    config.parser.set('db', 'user', 'postgres')
    config.parser.set('db', 'password', '')
    with open(config.path, 'w') as configfile:
        config.parser.write(configfile)


if __name__ == '__main__':
    create_travis_config()



# -*- coding: utf-8 -*-
"""
    The Efesto configuration module

    Copyright (C) 2016 Jacopo Cascioli

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import configparser
import os


class Config(object):

    def __init__(self, config_file='efesto.cfg'):
        self.path = self.find_path(config_file)
        self.parser = configparser.ConfigParser()
        self.parser.read(self.path)

    def find_path(self, path):
        """
        Finds the absolute path of the configuration file, looking in the
        current working directory (default) and its parent folder.

        The parent folder support is necessary for Sphinx.
        """
        possible_paths = [
            os.path.join(os.getcwd(), path),
            os.path.join(os.getcwd(), '..', path)
        ]
        for fullpath in possible_paths:
            if os.path.isfile(fullpath):
                return fullpath
        raise ValueError('The configuration file was not found at %s' % (path))



# -*- coding: utf-8 -*-
"""
    The Efesto base module.

    Copyright (C) 2016 Jacopo Cascioli

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from peewee import PostgresqlDatabase


from .Config import Config

config = Config()


db = PostgresqlDatabase(
    config.parser.get('db', 'name'),
    user=config.parser.get('db', 'user'),
    password=config.parser.get('db', 'password'),
    host=config.parser.get('db', 'host')
)



# -*- coding: utf-8 -*-
"""
    The models used by Efesto.

    Copyright (C) 2016 Jacopo Cascioli

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
from peewee import (BooleanField, CharField, DateTimeField, FloatField,
                    ForeignKeyField, IntegerField, PrimaryKeyField, TextField)
from playhouse.signals import Model, post_delete, pre_delete, pre_save


from .Base import db
from .Crypto import generate_hash, hexlify_


class Base(Model):
    """A base model that will use our Postgresql database"""
    class Meta:
        database = db
        validate_backrefs = False


class Users(Base):
    """
    Users.
    """
    id = PrimaryKeyField(primary_key=True)
    name = CharField(unique=True)
    email = CharField()
    password = CharField()
    rank = IntegerField()
    enabled = BooleanField()
    last_login = DateTimeField(null=True)

    def can(self, requested_action, item):
        if self.rank == 10:
            return True
        else:
            if requested_action == 'create':
                action = 'read'
            else:
                action = requested_action
            model_name = getattr(item._meta, 'db_table')
            rules = AccessRules.select()\
                .where((AccessRules.user == self.id) |
                       (AccessRules.rank == self.rank))\
                .where(AccessRules.model == model_name)\
                .where((AccessRules.item == None) |
                       (AccessRules.item == item.id))\
                .where(getattr(AccessRules, action) != None)\
                .order_by(AccessRules.level.desc(), AccessRules.item.asc(),
                          AccessRules.rank.desc())\
                .limit(1)
            if len(rules) > 0:
                if requested_action == 'create':
                    if getattr(rules[0], action) >= 5:
                        return True
                    else:
                        return False
                else:
                    if getattr(rules[0], action) >= 1:
                        return True
                    else:
                        return False
            return False


@pre_save(sender=Users)
def on_user_save(model_class, instance, created):
    """
    Hashes the password.
    """
    dirty = getattr(instance, '_dirty')
    if 'password' in dirty:
        instance.password = generate_hash(instance.password)


class Types(Base):
    """
    The Types specify the custom types that should be generated. Only enabled
    types will be generated.
    """
    id = PrimaryKeyField(primary_key=True)
    name = CharField(unique=True)
    enabled = BooleanField()


@post_delete(sender=Types)
def on_type_delete(model_class, instance):
    """
    Drops a type table when the type is deleted.
    """
    model = make_model(instance)
    model.drop_table()


@pre_delete(sender=Types)
def on_type_pre_delete(model_class, instance):
    """
    Checks whether a type has existing instances.
    """
    instance.enabled = 1
    instance.save()
    model = make_model(instance)
    if model.select().count() > 0:
        raise ValueError('This type has still existing instances')


class Fields(Base):
    """
    The fields are used to generate the columns for custom models/tables for
    the types specified in Types.

    name: the name of the field
    type: the type to which the field belongs
    unique: whether the generated column should be unique
    nullable: whether the generated column should have be nullable
    label: a label for the column, for informative purposes
    description: a description for the column, for infformative purposes
    """
    id = PrimaryKeyField(primary_key=True)
    name = CharField()
    type = ForeignKeyField(Types)
    field_type = CharField()
    unique = BooleanField(null=True)
    nullable = BooleanField(null=True)
    label = CharField(null=True)
    description = CharField(null=True)


class AccessRules(Base):
    """
    AccessRules define the permissions that an users or a group of users have
    on a single item or on a group of items.
    """
    id = PrimaryKeyField(primary_key=True)
    user = ForeignKeyField(Users, null=True)
    rank = IntegerField(null=True)
    item = IntegerField(null=True)
    model = CharField(null=True)
    level = IntegerField()
    read = IntegerField(null=True)
    edit = IntegerField(null=True)
    eliminate = IntegerField(null=True)


class EternalTokens(Base):
    """
    EternalTokens are server-stored tokens used for authentication purposes.
    Normally, Efesto would use client-stored tokens, but in some cases
    server-stored tokens are necessary.
    For example an application interacting with Efesto without requiring a
    final user's credentials (public data, password recovery, etc.)
    """
    id = PrimaryKeyField(primary_key=True)
    name = CharField()
    user = ForeignKeyField(Users)
    token = CharField()


@pre_save(sender=EternalTokens)
def on_token_save(model_class, instance, created):
    """
    Peewee hook that generates a random token whenever an eternal token is
    created.
    """
    dirty = getattr(instance, '_dirty')
    if 'token' in dirty:
        instance.token = hexlify_(os.urandom(24))


def make_model(custom_type):
    """
    Generates a model based on a Type entry, using the columns specified in
    Fields.
    """
    if custom_type.enabled:
        attributes = {}
        attributes['owner'] = ForeignKeyField(Users)
        fields_dict = {'string': TextField, 'int': IntegerField,
                       'float': FloatField, 'bool': BooleanField,
                       'date': DateTimeField}
        columns = Fields.select().where(Fields.type == custom_type.id)
        for column in columns:
            if column.field_type in fields_dict:
                args_dict = {}
                if column.nullable:
                    args_dict['null'] = True

                if column.unique:
                    args_dict['unique'] = True
                field_instance = fields_dict[column.field_type]
                attributes[column.name] = field_instance(**args_dict)
            else:
                parent_type = Types.get(Types.name == column.field_type)
                parent_model = make_model(parent_type)
                attributes[column.name] = ForeignKeyField(parent_model)
        model = type('%s' % (custom_type.name), (Base, ), attributes)
        db.create_tables([model], safe=True)
        return model
    else:
        raise ValueError('Cannot generate a model for a disabled type')



# -*- coding: utf-8 -*-
"""
    Efesto blueprints script.

    This script will load / dump blueprints.
"""
import argparse
import os

from efesto.Blueprints import dump_blueprint, load_blueprint


def load(file_to_load):
    path = os.path.join(os.getcwd(), file_to_load)
    if os.path.isfile(path):
        load_blueprint(path)
        print('Blueprint loaded successfully!')
    else:
        print('The provided path is not a file.')


def dump(dump_destination):
    path = os.path.join(os.getcwd(), dump_destination)
    dump_blueprint(path)
    print('Blueprint dumped successfully!')


def blueprints():
    parser = argparse.ArgumentParser()
    parser.add_argument('--dump', dest='dump')
    parser.add_argument('--load', dest='load')
    args = parser.parse_args()
    if args.load:
        load(args.load)

    if args.dump:
        dump(args.dump)



# -*- coding: utf-8 -*-
"""
   The Eefesto Resources module.

   Copyright (C) 2016 Jacopo Cascioli

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import json
from datetime import datetime
import falcon


from peewee import FieldDescriptor, ObjectIdDescriptor, RelationDescriptor
from .Auth import (authenticate_by_password, authenticate_by_token,
                   generate_token)
from .Base import config
from .Models import EternalTokens
from .Siren import hinder


def model_columns(model):
    columns = []
    for i in model.__dict__:
        if isinstance(model.__dict__[i], FieldDescriptor):
            if not isinstance(model.__dict__[i], RelationDescriptor):
                columns.append(i)

        if isinstance(model.__dict__[i], ObjectIdDescriptor):
            columns.append(i)
    return columns


def build_fields(request):
    fields = ['id']
    if '_fields' in request.params:
        if request.params['_fields'] != 'all':
            fields = request.params['_fields'].split(',')
            fields.append('id')
        else:
            fields = 'all'
    return fields


def build_query(model, params):
    query = model.select()
    for key, argument in params.items():
        if argument[0] == '<':
            query = query.where(getattr(model, key) <= argument[1:])
        elif argument[0] == '>':
            query = query.where(getattr(model, key) >= argument[1:])
        elif argument[0] == '!':
            query = query.where(getattr(model, key) != argument[1:])
        else:
            query = query.where(getattr(model, key) == argument)
    return query


def build_item(columns, fields, currrent_item):
    item = {}
    for column in columns:
        if column in fields or fields == 'all':
            item[column] = getattr(currrent_item, column)
    return item


def build_order(model, query, order):
    if order[0] == '<':
        query = query.order_by(getattr(model, order[1:]).desc())
    elif order[0] == '>':
        query = query.order_by(getattr(model, order[1:]).asc())
    else:
        query = query.order_by(getattr(model, order).asc())
    return query


def build_link_headers(request, response, count, items, page):
    domain = '{}://{}?page=%s&items={}'.format(request.protocol,
                                               request.host, items)
    last_page = int(count / items)

    if page != 1:
        prev_page = page - 1
        url = domain % (prev_page)
        response.add_link(url, rel='prev')

    if page != last_page:
        last_url = domain % (last_page)
        response.add_link(last_url, rel='last')
        next_page = page + 1
        if next_page != last_page:
            next_url = domain % (next_page)
            response.add_link(next_url, rel='next')


def item_to_dictionary(model, item):
    item_dict = {}
    for k in model.__dict__:
        if isinstance(model.__dict__[k], FieldDescriptor):
            if not isinstance(model.__dict__[k], RelationDescriptor):
                item_dict[k] = getattr(item, k)
        if isinstance(model.__dict__[k], ObjectIdDescriptor):
            item_dict[k] = getattr(item, k)
    return item_dict


def last_page(count, items):
    pages = int(count / items)
    if pages == 0:
        return 1
    return pages


def on_get(self, request, response):
    user = None
    if request.auth:
        user = authenticate_by_token(request.auth)

    if user is None:
        raise falcon.HTTPUnauthorized('Login required', 'Please login',
                                      ['Basic realm="Login Required"'])

    columns = model_columns(self.model)
    params = {}
    for i in columns:
        if i in request.params:
            params[i] = request.params[i]

    page = 1
    if 'page' in request.params:
        page = int(request.params['page'])

    items = 20
    if 'items' in request.params:
        items = int(request.params['items'])

    order = None
    if 'order_by' in request.params:
        order = request.params['order_by']

    fields = build_fields(request)

    query = build_query(self.model, params)
    if order is not None:
        query = build_order(self.model, query, order)
    count = query.count()

    body = []
    for i in query.paginate(page, items):
        if user.can('read', i):
            item = build_item(columns, fields, i)
            body.append(item)

    if len(body) == 0:
        response.status = falcon.HTTP_NO_CONTENT
        return response

    def json_serial(obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        raise TypeError('Type not serializable')
    pages = last_page(count, items)
    s = hinder(body, path=request.path, page=page, last_page=pages)
    response.body = json.dumps(s, default=json_serial)

    if count > items:
        build_link_headers(request, response, count, items, page)


def on_post(self, request, response):
    request._parse_form_urlencoded()
    user = None
    if request.auth:
        user = authenticate_by_token(request.auth)

    if user is None:
        raise falcon.HTTPUnauthorized('Login required', 'Please login',
                                      ['Basic realm="Login Required"'])

    if 'owner' not in request.params:
        request.params['owner'] = user.id
    new_item = self.model(**request.params)
    if user.can('read', new_item):
        new_item.save()
        response.status = falcon.HTTP_CREATED
        path = '{}/{}'.format(request.path, new_item.id)
        s = hinder(new_item.__dict__['_data'], path=path)
        response.body = json.dumps(s)
    else:
        raise falcon.HTTPForbidden('Forbidden access', 'You do not have the \
required permissions for this action')


def on_get_resource(self, request, response, id=0):
    user = None
    if request.auth:
        user = authenticate_by_token(request.auth)

    if user is None:
        raise falcon.HTTPUnauthorized('Login required', 'Please login',
                                      ['Basic realm="Login Required"'])

    try:
        item = self.model.get(getattr(self.model, 'id') == id)
    except:
        description = 'The resource you are looking for does not exist'
        raise falcon.HTTPNotFound(title='Not found', description=description)

    if user.can('read', item):
        item_dict = item_to_dictionary(self.model, item)

        def json_serial(obj):
            if isinstance(obj, datetime):
                return obj.isoformat()
            raise TypeError('Type not serializable')
        s = hinder(item_dict, path=request.path)
        response.body = json.dumps(s, default=json_serial)
    else:
        raise falcon.HTTPForbidden('Forbidden access', 'You do not have the \
required permissions for this action')


def on_patch_resource(self, request, response, id=0):
    user = None
    if request.auth:
        user = authenticate_by_token(request.auth)

    if user is None:
        raise falcon.HTTPUnauthorized('Login required', 'Please login',
                                      ['Basic realm="Login Required"'])

    try:
        item = self.model.get(getattr(self.model, 'id') == id)
    except:
        description = 'The resource you are looking for does not exist'
        raise falcon.HTTPNotFound(title='Not found', description=description)

    if user.can('edit', item):
        stream = request.stream.read().decode('UTF-8')
        if len(stream) > 0:
            for i in stream.split('&'):
                arg = i.split('=')
                setattr(item, arg[0], arg[1])
            item.save()

        item_dict = item_to_dictionary(self.model, item)
        s = hinder(item_dict, path=request.path)
        response.body = json.dumps(s)
    else:
        raise falcon.HTTPForbidden('Forbidden access', 'You do not have the \
required permissions for this action')


def on_delete_resource(self, request, response, id=0):
    user = None
    if request.auth:
        user = authenticate_by_token(request.auth)

    if user is None:
        raise falcon.HTTPUnauthorized('Login required', 'Please login',
                                      ['Basic realm="Login Required"'])

    try:
        item = self.model.get(getattr(self.model, 'id') == id)
    except:
        description = 'The resource you are looking for does not exist'
        raise falcon.HTTPNotFound(title='Not found', description=description)

    if user.can('eliminate', item):
        item.delete_instance()
        response.status = falcon.HTTP_NO_CONTENT
    else:
        raise falcon.HTTPForbidden('Forbidden access', 'You do not have the \
required permissions for this action')


def make_collection(model):
    """
    The make_collection function acts as generator of collection for models.
    """
    attributes = {
        'model': model,
        'on_get': on_get,
        'on_post': on_post
    }
    return type('mycollection', (object, ), attributes)


def make_resource(model):
    attributes = {
        'model': model,
        'on_get': on_get_resource,
        'on_patch': on_patch_resource,
        'on_delete': on_delete_resource
    }
    return type('mycollection', (object, ), attributes)


class TokensResource:
    """
    The TokensResource resource handles tokens requests.
    """
    def on_post(self, request, response):
        request._parse_form_urlencoded()
        if ('password' not in request.params or
                'username' not in request.params):
            raise falcon.HTTPBadRequest('Bad request',
                                        'A required parameter is missing')

        authentication = authenticate_by_password(request.params['username'],
                                                  request.params['password'])
        if authentication is None:
            raise falcon.HTTPForbidden('Forbidden access',
                                       'The credentials provided are invalid')

        if 'token_name' in request.params:
            try:
                t = EternalTokens.get(
                    EternalTokens.name == request.params['token_name'],
                    EternalTokens.user == authentication.id
                ).token
            except:
                raise falcon.HTTPForbidden('Forbidden access',
                                           'The credentials provided are \
invalid')
            token = generate_token(token=t)
        else:
            expiration = config.parser.getint('security', 'token_expiration')
            token = generate_token(expiration=expiration,
                                   user=request.params['username'])
        response.status = falcon.HTTP_OK
        response.body = json.dumps({'token': token})


class RootResource:
    """
    The RootResource resource handles requests made to the root endpoint.
    """
    def __init__(self, data):
        self.data = data

    def on_get(self, request, response):
        response.body = json.dumps(self.data)



# -*- coding: utf-8 -*-
"""
    The Efesto Siren module.

    Copyright (C) 2016 Jacopo Cascioli

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


def hinder(data, cls=None, path=None, page=None, last_page=None):
    siren = {}
    if type(data) == dict:
        siren['properties'] = data
        if path:
            siren['links'] = [{'rel': ['self'], 'href': path}]
    elif type(data) == list:
        siren['properties'] = {'count': len(data)}
        if path:
            if page == None:
                siren['links'] = [{'rel': ['self'], 'href': path}]
            elif page > 0:
                current_path = '{}?page={}'.format(path, page)
                if page == 1:
                    current_path = path
                siren['links'] = [{'rel': ['self'], 'href': current_path}]
                if page > 1:
                    prev_path = '{}?page={}'.format(path, page - 1)
                    prev = {'rel': ['previous'], 'href': prev_path}
                    siren['links'].append(prev)
                if last_page:
                    last_path = '{}?page={}'.format(path, last_page)
                    siren['links'].append({'rel': ['last'], 'href': last_path})
                    if last_page > page:
                        next_path = '{}?page={}'.format(path, page + 1)
                        next = {'rel': ['next'], 'href': next_path}
                        siren['links'].append(next)
                else:
                    next_path = '{}?page={}'.format(path, page + 1)
                    siren['links'].append({'rel': ['next'], 'href': next_path})

        entities = []
        for item in data:
            entity = {}
            entity['properties'] = item
            if path:
                href = '{}/{}'.format(path, item['id'])
                entity['href'] = href
            entity['rel'] = ['item']
            entities.append(entity)
        siren['entities'] = entities

    if cls:
        siren['class'] = [cls]
    return siren



# -*- coding: utf-8 -*-
"""
    The Efesto authentication module.

    Copyright (C) 2016 Jacopo Cascioli

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import base64
from itsdangerous import (JSONWebSignatureSerializer as Serializer,
                          TimedJSONWebSignatureSerializer as TimedSerializer)

from .Base import config
from .Crypto import compare_hash
from .Models import EternalTokens, Users


def generate_token(expiration=600, **kwargs):
    s = TimedSerializer(config.parser.get('security', 'secret'),
                        expires_in=expiration)
    return s.dumps(kwargs).decode('UTF-8')


def read_token(token):
    """
    Reads a token. If the token contains a user parameter, the token is read
    with TimedSerializer.
    """
    s = Serializer(config.parser.get('security', 'secret'))
    result = s.loads(token)
    if 'user' in result:
        t = TimedSerializer(config.parser.get('security', 'secret'))
        return t.loads(token)
    else:
        return result


def authenticate_by_password(username, password):
    """
    Authenticates a user by username and password. Usually this occurs only
    when an user needs a token.
    """
    try:
        user = Users.get(Users.name == username, Users.enabled == True)
    except:
        return None

    if compare_hash(password, user.password):
        return user


def parse_auth_header(auth_string):
    """
    Parses a basic auth header.
    """
    return base64.b64decode(auth_string.split()[1]).decode('latin-1')


def authenticate_by_token(auth_header):
    """
    Authenticates a user with a token or an eternal token.
    """
    try:
        parsed = parse_auth_header(auth_header)
        auth_dict = read_token(parsed.split(':')[1])
    except:
        return None

    if 'token' in auth_dict:
        auth_token = auth_dict['token']
        token = EternalTokens.get(EternalTokens.token == auth_token)
        if token.user.enabled == True:
            return token.user
        return None
    else:
        try:
            user = Users.get(Users.name == auth_dict['user'],
                             Users.enabled == True)
        except:
            user = None
        return user



# -*- coding: utf-8 -*-
"""
    Efesto quickstart script.

    This script will set up Efesto, creating the tables, the secret and the
    administrator account.
"""
import argparse
import getpass
import os
import sys
from binascii import hexlify

from colorama import Fore, Style
import efesto
from efesto.Base import config, db
from efesto.Models import AccessRules, EternalTokens, Fields, Types, Users
from peewee import OperationalError, ProgrammingError


sys.path.insert(0, '')


def message(message, colour):
    """
    Prints a coloured message using colorama.
    """
    text_colours = {'green': Fore.GREEN, 'red': Fore.RED, 'blue': Fore.BLUE}
    print(text_colours[colour] + message + Style.RESET_ALL)


def create_tables():
    """
    Creates the tables.
    """
    try:
        db.create_tables([Users, Types, Fields, AccessRules, EternalTokens])
        message('Tables created!', 'green')
    except OperationalError:
        message(('An error occured during tables creation. '
                 'Please check your database credentials.'), 'red')
        exit()
    except ProgrammingError:
        message(('An error occurred during tables creation. '
                 'Please check your database.'), 'red')
        exit()
    except:
        message('An unknown error occurred during tables creation', 'red')
        exit()


def create_config():
    secret_key = str(hexlify(os.urandom(24)), encoding='utf8')
    config.parser.set('main', 'installed', 'True')
    config.parser.set('security', 'secret', secret_key)
    with open(config.path, 'w') as configfile:
        config.parser.write(configfile)
        message('Configuration successfully updated!', 'green')


def create_admin():
    """
    Creates the default administrator.
    """
    print('Efesto will now create the default administrator')
    admin_name = input('Administrator name: ')
    admin_email = input('Administrator email: ')
    admin_password = getpass.getpass('Administrator password: ')
    admin = Users(name=admin_name, email=admin_email, rank=10,
                  password=admin_password, enabled=True)
    admin.save()
    print(Fore.GREEN + 'Admin created!')


def quickstart():
    parser = argparse.ArgumentParser()
    parser.add_argument('--no-admin', dest='admin', action='store_false')
    parser.set_defaults(admin=True)
    args = parser.parse_args()
    installed = config.parser.getboolean('main', 'installed')
    if installed != True:
        message('This script will help you setup Efesto', 'blue')
        version_message = 'Efesto version: %s' % (efesto.__version__)
        message(version_message, 'blue')
        create_tables()
        create_config()
        if args.admin:
            create_admin()
        message('Set up successful.', 'green')
    else:
        message('It seems Efesto has been already set up!', 'red')


if __name__ == '__main__':
    quickstart()



# -*- coding: utf-8 -*-
"""
    Crypto
    ~~~~~~

    This module implements cryptographic functions in efesto.

    See:
    https://github.com/mitsuhiko/python-pbkdf2 (original)
    https://github.com/reedobrien/python-pbkdf2 (Python3 compatible)

    :copyright: (c) Copyright 2011 by Armin Ronacher.
    :license: BSD
"""
import hashlib
import hmac
import os
from binascii import hexlify
from itertools import starmap
from operator import xor
from struct import Struct


from .Base import config


_pack_int = Struct('>I').pack


def bytes_(s, encoding='utf8', errors='strict'):
    if isinstance(s, str):
        return s.encode(encoding, errors)
    return s


def hexlify_(s):
    return str(hexlify(s), encoding='utf8')


def range_(*args):
    return range(*args)


def pbkdf2_hex(data, salt, iterations=1000, keylen=24, hashfunc=None):
    """Like :func:`pbkdf2_bin` but returns a hex encoded string."""
    return hexlify_(pbkdf2_bin(data, salt, iterations, keylen, hashfunc))


def pbkdf2_bin(data, salt, iterations=1000, keylen=24, hashfunc=None):
    """Returns a binary digest for the PBKDF2 hash algorithm of `data`
    with the given `salt`.  It iterates `iterations` time and produces a
    key of `keylen` bytes.  By default SHA-1 is used as hash function,
    a different hashlib `hashfunc` can be provided.
    """
    hashfunc = hashfunc or hashlib.sha1
    mac = hmac.new(bytes_(data), None, hashfunc)

    def _pseudorandom(x, mac=mac):
        h = mac.copy()
        h.update(bytes_(x))
        return [x for x in h.digest()]
    buf = []
    for block in range_(1, -(-keylen // mac.digest_size) + 1):
        rv = u = _pseudorandom(bytes_(salt) + _pack_int(block))
        for i in range_(iterations - 1):
            u = _pseudorandom(bytes(u))
            rv = starmap(xor, zip(rv, u))
        buf.extend(rv)
    return bytes(buf)[:keylen]


def safe_str_cmp(a, b):
    if len(a) != len(b):
        return False
    rv = 0
    for x, y in zip(a, b):
        rv |= ord(x) ^ ord(y)
    return rv == 0


def generate_hash(string_to_hash):
    """ Generates a complete hash, that includes algorithm, iterations, salt
    and hash separated by hash_separator.

    E.g. 'PBKDF2$iterations$salt$hashed_string'
    """
    salt_length = config.parser.getint('security', 'salt_length')
    iterations = config.parser.getint('security', 'iterations')
    key_length = config.parser.getint('security', 'key_length')
    salt = hexlify_(os.urandom(salt_length))
    hashed_string = pbkdf2_hex(string_to_hash, salt, iterations=iterations,
                               keylen=key_length)
    return '$'.join(['PBKDF2-256', str(iterations), salt, hashed_string])


def compare_hash(plain_string, old_hash):
    """ Compares an hash generated with plain_string to an exisiting hash."""
    key_length = config.parser.getint('security', 'key_length')
    splitted = old_hash.split('$')
    iterations = int(splitted[1])
    salt = splitted[2]
    new_hash = pbkdf2_hex(plain_string, salt, iterations=iterations,
                          keylen=key_length)
    return safe_str_cmp(new_hash, splitted[3])



#!/usr/bin/env python
from setuptools import find_packages, setup


setup(
    name='efesto',
    description='Efesto is a RESTful API generator based on the Falcon.',
    url='https://github.com/vesuvium/efesto',
    author='Jacopo Cascioli',
    author_email='jacopocascioli@gmail.com',
    license='GPL3',
    version='0.6.1',
    packages=find_packages(),
    tests_require=[
        'pytest',
        'pytest-falcon'
    ],
    setup_requires=['pytest-runner'],
    install_requires=[
        'falcon==1.0.0',
        'psycopg2>=2.6.1',
        'peewee>=2.8.0',
        'itsdangerous>=0.24',
        'colorama>=0.3.3'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Topic :: Internet :: WWW/HTTP',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    ],
    entry_points='''
        [console_scripts]
        efesto-quickstart=efesto.scripts.quickstart:quickstart
        efesto-blueprints=efesto.scripts.blueprints:blueprints
    '''
)
